@@ -170,7 +170,9 @@ void ipmi_fru_put_node(ipmi_fru_node_t *node);
  * the individual elements are not named.  The "index" is an index
  * into the array itself, starting at 0.  The elements of an array
  * node will always set the "name" to NULL.  If you fetch beyond the
- * end of an array, it will return EINVAL.
+ * end of an array, it will return EINVAL.  If you fetch a field and
+ * it returns a SUB_NODE type and "intval" is not -1, the sub_node
+ * field will be set to an array subnode.
  *
  * A record node is a node that has a set of elements that are named,
  * the "name" will always be set to a value when fetching one of
@@ -178,7 +180,9 @@ void ipmi_fru_put_node(ipmi_fru_node_t *node);
  * that, unlike arrays, individual fields in a record node may not be
  * present.  Fetching these fields will return ENOSYS, but there may
  * be valid fields after this.  This returns EINVAL if you go beyond
- * the last field.
+ * the last field.  If you fetch a field and it returns a SUB_NODE
+ * type and "intval" is -1, the sub_node field will be set to an
+ * record subnode.
  *
  * The various dtypes are:
  *    IPMI_FRU_DATA_INT  - sets "intval"
@@ -211,6 +215,55 @@ int ipmi_fru_node_get_field(ipmi_fru_node_t           *node,
 			    unsigned int              *data_len,
 			    ipmi_fru_node_t           **sub_node);
 
+/*
+ * Set an index in a node.  See the information for ipmi_fru_node_get_field
+ * above for information on the types and data values.
+ *
+ * This function returns EPERM if the field is not writable or EINVAL if
+ * something is invalid about the data or index or data type.
+ *
+ * Setting a base value (int, float, time, etc.) will set that value
+ * for the field.  Nothing magical or suprising.
+ *
+ * Setting an index that returns an array will insert a new value into
+ * the array at the given index (if the index is positive) or delete
+ * the value at -index-1 (if the index is negative).  To extend an
+ * array, write to an index beyond the current length and it will be
+ * automatically extended by one value.  The new element will be set
+ * to some initial value.
+ *
+ * Setting a record type subnode is not supported.
+ */
+int ipmi_fru_node_set_field(ipmi_fru_node_t           *node,
+			    unsigned int              index,
+			    enum ipmi_fru_data_type_e dtype,
+			    int                       intval,
+			    time_t                    time,
+			    double                    floatval,
+			    char                      *data,
+			    unsigned int              data_len);
+
+/*
+ * Can an index in a node be set?  Returns 0 if so, an error if not.
+ */
+int ipmi_fru_node_settable(ipmi_fru_node_t *node,
+			   unsigned int    index);
+
+/* Only for arrays, get's the array's subtype. */
+int ipmi_fru_node_get_subtype(ipmi_fru_node_t           *node,
+			      enum ipmi_fru_data_type_e *subtype);
+
+/* Used to map enum-type values, where there is a limited set of value
+   values.  Pass in -1 for "pos" to get the first value, "pos" will be
+   set to the actual value.  nextpos will be set to the next valid
+   value.  data will return a pointer to an allocated string for the
+   value.  For integer values, pos will map to the actual intval
+   values.  For others, the order is undetermined. */
+int ipmi_fru_node_get_enum_val(ipmi_fru_node_t *node,
+			       unsigned int    index,
+			       int             *pos,
+			       int             *nextpos,
+			       const char      **data);
 
 /* Free data that comes from ipmi_fru_node_get_field if the data return is
    non-NULL. */
@@ -415,6 +468,12 @@ int ipmi_fru_get_multi_record_data(ipmi_fru_t    *fru,
 				   unsigned int  num,
 				   unsigned char *data,
 				   unsigned int  *length);
+/* Get part of the multirecord data. */
+int ipmi_fru_get_multi_record_slice(ipmi_fru_t    *fru,
+				    unsigned int  num,
+				    unsigned int  offset,
+				    unsigned int  length,
+				    unsigned char *data);
 
 /*
  * This interface lets you get the FRU data by name.
@@ -505,6 +564,8 @@ int ipmi_fru_str_to_index(char *name);
 /* Note that the length for adding multi-records is ignored, it will
    calculate it to the end of the data. */
 
+/* If you set the length to zero when adding an area, you will get a
+   minimally sized area. */
 int ipmi_fru_add_area(ipmi_fru_t   *fru,
 		      unsigned int area,
 		      unsigned int offset,
@@ -629,12 +690,48 @@ int ipmi_fru_set_product_info_custom(ipmi_fru_t   *fru,
 				     char         *str,
 				     unsigned int len);
 
+/* Set the type field for a multi-record. */
+int ipmi_fru_set_multi_record_type(ipmi_fru_t    *fru,
+				   unsigned int  num,
+				   unsigned char type);
+/* Replace the data in a multi-record.  "data" may not be NULL. */
+int ipmi_fru_set_multi_record_data(ipmi_fru_t    *fru,
+				   unsigned int  num,
+				   unsigned char *data,
+				   unsigned int  length);
+/* If "data" is non-NULL, overwrite an existing multirecord with new
+   information, or insert a new record at the end if "num" is the same
+   or larger then the current number of multirecords.  If "data" is
+   NULL, then delete the given multirecord.  For deletion, type and
+   version are ignored. */
 int ipmi_fru_set_multi_record(ipmi_fru_t    *fru,
 			      unsigned int  num,
 			      unsigned char type,
 			      unsigned char version,
 			      unsigned char *data,
 			      unsigned int  length);
+/* Overwrite data in the given multirecord's data with new
+   information.  Overwrite data at the given offset to the given
+   length.  The length must be within the multirecord's current
+   length. */
+int ipmi_fru_ovw_multi_record_data(ipmi_fru_t    *fru,
+				   unsigned int  num,
+				   unsigned char *data,
+				   unsigned int  offset,
+				   unsigned int  length);
+/* Insert new data in the given multirecord.  The total length may not
+   exceed the maximum length of a multirecord. */
+int ipmi_fru_ins_multi_record_data(ipmi_fru_t    *fru,
+				   unsigned int  num,
+				   unsigned char *data,
+				   unsigned int  offset,
+				   unsigned int  length);
+/* Delete data in the given multirecord.  The total length may not
+   go below zero. */
+int ipmi_fru_del_multi_record_data(ipmi_fru_t    *fru,
+				   unsigned int  num,
+				   unsigned int  offset,
+				   unsigned int  length);
 
 /*
  * A generic interface for setting values by index.  The function to use
