@@ -236,17 +236,37 @@ typedef struct atca_site_s
     unsigned char site_number;
 } atca_site_t;
 
+#define MAX_EMU_ADDR		16
+#define MAX_EMU_ADDR_DATA	64
+typedef struct emu_addr_s
+{
+    unsigned char valid;
+    unsigned char addr_type;
+    unsigned char addr_data[MAX_EMU_ADDR_DATA];
+    unsigned int  addr_len;
+} emu_addr_t;
+
 struct emu_data_s
 {
     int        bmc_mc;
     lmc_data_t *ipmb[128];
 
-    int         atca_mode;
-    atca_site_t atca_sites[128]; /* Indexed by HW address. */
+    int          atca_mode;
+    atca_site_t  atca_sites[128]; /* Indexed by HW address. */
+    uint32_t     atca_fru_inv_curr_timestamp;
+    uint16_t     atca_fru_inv_curr_lock_id;
+    int          atca_fru_inv_locked;
+    int          atca_fru_inv_lock_timeout;
+
+    unsigned char *temp_fru_inv_data;
+    unsigned int  temp_fru_inv_data_len;
 
     void *user_data;
 
     ipmi_emu_sleep_cb sleeper;
+
+    struct timeval last_addr_change_time;
+    emu_addr_t addr[MAX_EMU_ADDR];
 };
 
 static void picmg_led_set(lmc_data_t *mc, sensor_t *sensor);
@@ -261,6 +281,19 @@ static void picmg_led_set(lmc_data_t *mc, sensor_t *sensor);
 #define IPMI_DEVID_SDR_REPOSITORY_DEV	(1 << 1)
 #define IPMI_DEVID_SENSOR_DEV		(1 << 0)
 
+void
+ipmi_emu_tick(emu_data_t *emu, unsigned int seconds)
+{
+    if (emu->atca_fru_inv_locked) {
+	emu->atca_fru_inv_lock_timeout -= seconds;
+	if (emu->atca_fru_inv_lock_timeout < 0) {
+	    emu->atca_fru_inv_locked = 0;
+	    free(emu->temp_fru_inv_data);
+	    emu->temp_fru_inv_data = NULL;
+	}
+    }
+}
+
 /*
  * SEL handling commands.
  */
@@ -2872,7 +2905,7 @@ handle_picmg_get_properties(lmc_data_t    *mc,
 {
     rdata[0] = 0;
     rdata[1] = IPMI_PICMG_GRP_EXT;
-    rdata[2] = 0x02; /* Version 2.0 */
+    rdata[2] = 0x22; /* Version 2.2 */
     rdata[3] = 0; /* Only have one FRU. */
     rdata[4] = 0; /* As defined by spec. */
     *rdata_len = 5;
@@ -2969,7 +3002,7 @@ handle_picmg_cmd_fru_control(lmc_data_t    *mc,
 	return;
     }
 
-    if (msg->data[2] != 0) {
+    if (msg->data[2] >= 4) {
 	rdata[0] = IPMI_INVALID_DATA_FIELD_CC;
 	*rdata_len = 1;
 	return;
@@ -3420,6 +3453,308 @@ handle_picmg_cmd_bused_resource(lmc_data_t    *mc,
     handle_invalid_cmd(mc, rdata, rdata_len);
 }
 
+static void
+handle_picmg_cmd_ipmb_link_info(lmc_data_t    *mc,
+				ipmi_msg_t    *msg,
+				unsigned char *rdata,
+				unsigned int  *rdata_len)
+{
+    handle_invalid_cmd(mc, rdata, rdata_len);
+}
+
+static void
+handle_picmg_cmd_shelf_power_allocation(lmc_data_t    *mc,
+					ipmi_msg_t    *msg,
+					unsigned char *rdata,
+					unsigned int  *rdata_len)
+{
+    if (check_msg_length(msg, 2, rdata, rdata_len))
+	return;
+
+    if (msg->data[1] > 1) {
+	rdata[0] = IPMI_DESTINATION_UNAVAILABLE_CC;
+	*rdata_len = 1;
+	return;
+    }
+
+    rdata[0] = 0;
+    rdata[1] = IPMI_PICMG_GRP_EXT;
+    ipmi_set_uint16(rdata+2, 0);
+    if (msg->data[1] == 0) {
+	ipmi_set_uint16(rdata+4, 105);
+	ipmi_set_uint16(rdata+6, 227);
+	*rdata_len = 8;
+    } else {
+	ipmi_set_uint16(rdata+4, 227);
+	*rdata_len = 6;
+    }
+}
+
+static void
+handle_picmg_cmd_shelf_manager_ipmb_address(lmc_data_t    *mc,
+					    ipmi_msg_t    *msg,
+					    unsigned char *rdata,
+					    unsigned int  *rdata_len)
+{
+    handle_invalid_cmd(mc, rdata, rdata_len);
+}
+
+static void
+handle_picmg_cmd_set_fan_policy(lmc_data_t    *mc,
+				ipmi_msg_t    *msg,
+				unsigned char *rdata,
+				unsigned int  *rdata_len)
+{
+    handle_invalid_cmd(mc, rdata, rdata_len);
+}
+
+static void
+handle_picmg_cmd_get_fan_policy(lmc_data_t    *mc,
+				ipmi_msg_t    *msg,
+				unsigned char *rdata,
+				unsigned int  *rdata_len)
+{
+    handle_invalid_cmd(mc, rdata, rdata_len);
+}
+
+
+static void
+handle_picmg_cmd_fru_control_capabilities(lmc_data_t    *mc,
+					  ipmi_msg_t    *msg,
+					  unsigned char *rdata,
+					  unsigned int  *rdata_len)
+{
+    if (check_msg_length(msg, 2, rdata, rdata_len))
+	return;
+
+    if (msg->data[1] != 0) {
+	rdata[0] = IPMI_DESTINATION_UNAVAILABLE_CC;
+	*rdata_len = 1;
+	return;
+    }
+
+    rdata[0] = 0;
+    rdata[1] = IPMI_PICMG_GRP_EXT;
+    rdata[2] = 0x0e;
+    *rdata_len = 3;
+}
+
+static void
+handle_picmg_cmd_fru_inventory_device_lock_control(lmc_data_t    *mc,
+						   ipmi_msg_t    *msg,
+						   unsigned char *rdata,
+						   unsigned int  *rdata_len)
+{
+    emu_data_t *emu = mc->emu;
+    uint16_t   lock_id;
+
+    if (mc->ipmb != 0x20) {
+	handle_invalid_cmd(mc, rdata, rdata_len);
+	return;
+    }
+
+    if (check_msg_length(msg, 5, rdata, rdata_len))
+	return;
+
+    if (msg->data[1] != 254) {
+	rdata[0] = IPMI_DESTINATION_UNAVAILABLE_CC;
+	*rdata_len = 1;
+	return;
+    }
+
+    rdata[0] = 0;
+    rdata[1] = IPMI_PICMG_GRP_EXT;
+
+    switch (msg->data[2]) {
+    case 0:
+	rdata[2] = 0;
+	rdata[3] = 0;
+	ipmi_set_uint32(rdata+4, emu->atca_fru_inv_curr_timestamp);
+	*rdata_len = 8;
+	break;
+
+    case 1:
+	if (emu->atca_fru_inv_locked) {
+	    rdata[0] = 0x81;
+	    *rdata_len = 1;
+	    break;
+	}
+	if (mc->frus[254].length == 0) {
+	    rdata[0] = IPMI_NOT_SUPPORTED_IN_PRESENT_STATE_CC;
+	    *rdata_len = 1;
+	    break;
+	}
+	emu->temp_fru_inv_data = malloc(mc->frus[254].length);
+	if (!emu->temp_fru_inv_data) {
+	    rdata[0] = IPMI_OUT_OF_SPACE_CC;
+	    *rdata_len = 1;
+	    break;
+	}
+	emu->temp_fru_inv_data_len = mc->frus[254].length;
+	memcpy(emu->temp_fru_inv_data, mc->frus[254].data, 
+	       emu->temp_fru_inv_data_len);
+
+	emu->atca_fru_inv_locked = 1;
+	emu->atca_fru_inv_curr_lock_id++;
+	ipmi_set_uint16(rdata+2, emu->atca_fru_inv_curr_lock_id);
+	ipmi_set_uint32(rdata+4, emu->atca_fru_inv_curr_timestamp);
+	*rdata_len = 8;
+	emu->atca_fru_inv_lock_timeout = 20;
+	break;
+
+    case 2:
+	lock_id = ipmi_get_uint16(msg->data+3);
+	if (!emu->atca_fru_inv_locked
+	    || (lock_id != emu->atca_fru_inv_curr_lock_id))
+	{
+	    rdata[0] = 0x81;
+	    *rdata_len = 1;
+	    break;
+	}
+	emu->atca_fru_inv_locked = 0;
+	rdata[2] = 0;
+	rdata[3] = 0;
+	ipmi_set_uint32(rdata+4, emu->atca_fru_inv_curr_timestamp);
+	*rdata_len = 8;
+	free(emu->temp_fru_inv_data);
+	emu->temp_fru_inv_data = NULL;
+	break;
+
+    case 3:
+	lock_id = ipmi_get_uint16(msg->data+3);
+	if (!emu->atca_fru_inv_locked
+	    || (lock_id != emu->atca_fru_inv_curr_lock_id))
+	{
+	    rdata[0] = 0x81;
+	    *rdata_len = 1;
+	    break;
+	}
+	emu->atca_fru_inv_locked = 0;
+	rdata[2] = 0;
+	rdata[3] = 0;
+	ipmi_set_uint32(rdata+4, emu->atca_fru_inv_curr_timestamp);
+	*rdata_len = 8;
+	emu->atca_fru_inv_curr_timestamp++;
+	/* FIXME - validate data. */
+	memcpy(mc->frus[254].data, emu->temp_fru_inv_data,
+	       emu->temp_fru_inv_data_len);
+	free(emu->temp_fru_inv_data);
+	emu->temp_fru_inv_data = NULL;
+	break;
+
+    default:
+	rdata[0] = IPMI_INVALID_DATA_FIELD_CC;
+	*rdata_len = 1;
+	break;
+    }
+}
+
+static void
+handle_picmg_cmd_fru_inventory_device_write(lmc_data_t    *mc,
+					    ipmi_msg_t    *msg,
+					    unsigned char *rdata,
+					    unsigned int  *rdata_len)
+{
+    emu_data_t   *emu = mc->emu;
+    uint16_t     lock_id;
+    unsigned int offset;
+    unsigned int count;
+
+    if (mc->ipmb != 0x20) {
+	handle_invalid_cmd(mc, rdata, rdata_len);
+	return;
+    }
+
+    if (check_msg_length(msg, 6, rdata, rdata_len))
+	return;
+
+    if (msg->data[1] != 254) {
+	rdata[0] = IPMI_DESTINATION_UNAVAILABLE_CC;
+	*rdata_len = 1;
+	return;
+    }
+
+    lock_id = ipmi_get_uint16(msg->data+2);
+    if (!emu->atca_fru_inv_locked
+	|| (lock_id != emu->atca_fru_inv_curr_lock_id))
+    {
+	rdata[0] = 0x80;
+	*rdata_len = 1;
+	return;
+    }
+
+    /* Reset the timer. */
+    emu->atca_fru_inv_lock_timeout = 20;
+
+    offset = ipmi_get_uint16(msg->data+4);
+    count = msg->data_len - 6;
+
+    if (offset >= emu->temp_fru_inv_data_len) {
+	rdata[0] = IPMI_PARAMETER_OUT_OF_RANGE_CC;
+	*rdata_len = 1;
+	return;
+    }
+
+    if ((offset+count) > emu->temp_fru_inv_data_len) {
+	/* Too much data to put into FRU. */
+	rdata[0] = IPMI_REQUESTED_DATA_LENGTH_EXCEEDED_CC;
+	*rdata_len = 1;
+	return;
+    }
+
+    memcpy(emu->temp_fru_inv_data+offset, msg->data+6, count);
+
+    rdata[0] = 0;
+    rdata[1] = IPMI_PICMG_GRP_EXT;
+    rdata[2] = count;
+    *rdata_len = 3;
+}
+
+static void
+handle_picmg_cmd_get_shelf_manager_ip_addresses(lmc_data_t    *mc,
+						ipmi_msg_t    *msg,
+						unsigned char *rdata,
+						unsigned int  *rdata_len)
+{
+    emu_data_t   *emu = mc->emu;
+    unsigned int addr;
+    unsigned int count;
+    emu_addr_t   *ap = NULL;
+    int          i;
+
+    if (check_msg_length(msg, 2, rdata, rdata_len))
+	return;
+
+    addr = msg->data[1];
+    
+    for (count=0, i=0; i<MAX_EMU_ADDR; i++) {
+	if (emu->addr[i].valid) {
+	    if (count == addr)
+		ap = &(emu->addr[i]);
+	    count++;
+	}
+    }
+
+    if (addr >= count) {
+	rdata[0] = IPMI_PARAMETER_OUT_OF_RANGE_CC;
+	*rdata_len = 1;
+	return;
+    }
+
+    rdata[0] = 0;
+    ipmi_set_uint32(rdata+1, emu->last_addr_change_time.tv_sec);
+    rdata[5] = count;
+    rdata[6] = 0x03;
+    rdata[7] = addr - 1;
+    rdata[8] = 20;
+
+    rdata[9] = ap->addr_type;
+    if (addr == 0)
+	rdata[9] |= 0x80;
+    memcpy(rdata+10, ap->addr_data, ap->addr_len);
+    *rdata_len = 10 + ap->addr_len;
+}
+
 int
 ipmi_emu_atca_enable(emu_data_t *emu)
 {
@@ -3547,6 +3882,44 @@ handle_picmg_msg(lmc_data_t    *mc,
 	handle_picmg_cmd_bused_resource(mc, msg, rdata, rdata_len);
 	break;
 
+    case IPMI_PICMG_CMD_IPMB_LINK_INFO:
+	handle_picmg_cmd_ipmb_link_info(mc, msg, rdata, rdata_len);
+	break;
+      
+    case IPMI_PICMG_CMD_SHELF_POWER_ALLOCATION:
+	handle_picmg_cmd_shelf_power_allocation(mc, msg, rdata, rdata_len);
+	break;
+
+    case IPMI_PICMG_CMD_SHELF_MANAGER_IPMB_ADDRESS:
+	handle_picmg_cmd_shelf_manager_ipmb_address(mc, msg, rdata, rdata_len);
+	break;
+
+    case IPMI_PICMG_CMD_SET_FAN_POLICY:
+	handle_picmg_cmd_set_fan_policy(mc, msg, rdata, rdata_len);
+	break;
+
+    case IPMI_PICMG_CMD_GET_FAN_POLICY:
+	handle_picmg_cmd_get_fan_policy(mc, msg, rdata, rdata_len);
+	break;
+
+    case IPMI_PICMG_CMD_FRU_CONTROL_CAPABILITIES:
+	handle_picmg_cmd_fru_control_capabilities(mc, msg, rdata, rdata_len);
+	break;
+
+    case IPMI_PICMG_CMD_FRU_INVENTORY_DEVICE_LOCK_CONTROL:
+	handle_picmg_cmd_fru_inventory_device_lock_control(mc, msg, rdata,
+							   rdata_len);
+	break;
+
+    case IPMI_PICMG_CMD_FRU_INVENTORY_DEVICE_WRITE:
+	handle_picmg_cmd_fru_inventory_device_write(mc, msg, rdata, rdata_len);
+	break;
+
+    case IPMI_PICMG_CMD_GET_SHELF_MANAGER_IP_ADDRESSES:
+	handle_picmg_cmd_get_shelf_manager_ip_addresses(mc, msg, rdata,
+							rdata_len);
+	break;
+
     default:
 	handle_invalid_cmd(mc, rdata, rdata_len);
 	break;
@@ -3708,6 +4081,41 @@ ipmi_emu_alloc(void *user_data, ipmi_emu_sleep_cb sleeper)
     return data;
 }
 
+int
+ipmi_emu_set_addr(emu_data_t *emu, unsigned int addr_num,
+		  unsigned char addr_type,
+		  void *addr_data, unsigned int addr_len)
+{
+    emu_addr_t *addr;
+
+    if (addr_num >= MAX_EMU_ADDR)
+	return EINVAL;
+
+    addr = &(emu->addr[addr_num]);
+    if (addr_len > sizeof(addr->addr_data))
+	return EINVAL;
+
+    gettimeofday(&emu->last_addr_change_time, NULL);
+    addr->addr_type = addr_type;
+    memcpy(addr->addr_data, addr_data, addr_len);
+    addr->addr_len = addr_len;
+    addr->valid = 1;
+    return 0;
+}
+
+int
+ipmi_emu_clear_addr(emu_data_t *emu, unsigned int addr_num)
+{
+    emu_addr_t *addr;
+
+    if (addr_num >= MAX_EMU_ADDR)
+	return EINVAL;
+
+    addr = &(emu->addr[addr_num]);
+    addr->valid = 0;
+    return 0;
+}
+
 void
 ipmi_emu_sleep(emu_data_t *emu, struct timeval *time)
 {
