@@ -86,12 +86,12 @@ ipmi_format_msg(ipmi_con_t        *ipmi,
 	ipmi_system_interface_addr_t *si_addr
 	    = (ipmi_system_interface_addr_t *) addr;
 
-	if ((msg->data_len + 7) > *out_data_len)
+	if ((unsigned int) (msg->data_len + 7) > *out_data_len)
 	    return E2BIG;
 	if (ipmi->hacks & IPMI_CONN_HACK_20_AS_MAIN_ADDR)
 	    tmsg[0] = 0x20;
 	else
-	    tmsg[0] = ipmi->ipmb_addr; /* To the BMC. */
+	    tmsg[0] = ipmi->ipmb_addr[0]; /* To the BMC. */
 	tmsg[1] = (msg->netfn << 2) | si_addr->lun;
 	tmsg[2] = ipmb_checksum(tmsg, 2);
 	tmsg[3] = 0x81; /* Remote console IPMI Software ID */
@@ -107,20 +107,23 @@ ipmi_format_msg(ipmi_con_t        *ipmi,
 	ipmi_ipmb_addr_t *ipmb_addr = (ipmi_ipmb_addr_t *) addr;
 	int              do_broadcast = 0;
 
+	if (ipmb_addr->channel >= MAX_IPMI_USED_CHANNELS)
+	    return EINVAL;
+
 	if ((addr->addr_type == IPMI_IPMB_BROADCAST_ADDR_TYPE)
 	    && (!ipmi->broadcast_broken))
 	{
 	    do_broadcast = 1;
 	}
 
-	if ((msg->data_len + 15 + do_broadcast) > *out_data_len)
+	if ((unsigned int) (msg->data_len + 15 + do_broadcast) > *out_data_len)
 	    return E2BIG;
 
 	pos = 0;
 	if (ipmi->hacks & IPMI_CONN_HACK_20_AS_MAIN_ADDR)
 	    tmsg[pos++] = 0x20;
 	else
-	    tmsg[pos++] = ipmi->ipmb_addr; /* BMC is the bridge. */
+	    tmsg[pos++] = ipmi->ipmb_addr[0]; /* BMC is the bridge. */
 	tmsg[pos++] = (IPMI_APP_NETFN << 2) | 0;
 	tmsg[pos++] = ipmb_checksum(tmsg, 2);
 	tmsg[pos++] = 0x81; /* Remote console IPMI Software ID */
@@ -135,7 +138,7 @@ ipmi_format_msg(ipmi_con_t        *ipmi,
 	tmsg[pos++] = (msg->netfn << 2) | ipmb_addr->lun;
 	tmsg[pos++] = ipmb_checksum(tmsg+msgstart, 2);
 	msgstart = pos;
-	tmsg[pos++] = ipmi->ipmb_addr;
+	tmsg[pos++] = ipmi->ipmb_addr[ipmb_addr->channel];
 	tmsg[pos++] = (seq << 2) | 2; /* add 2 as the SMS LUN */
 	tmsg[pos++] = msg->cmd;
 	memcpy(tmsg+pos, msg->data, msg->data_len);
@@ -189,6 +192,8 @@ ipmi_handle_recv(ipmi_con_t    *ipmi,
     unsigned int  addr_len;
     unsigned int  seq;
     unsigned char *tmsg = data;
+    int           chan;
+    int           to_ret = 0;
 
     if (data_len < 8) { /* Minimum size of an IPMI msg. */
 	if (DEBUG_RAWMSG || DEBUG_MSG_ERR)
@@ -202,6 +207,14 @@ ipmi_handle_recv(ipmi_con_t    *ipmi,
 
     seq = data[4] >> 2;
 
+    if ((orig_addr->addr_type == IPMI_IPMB_BROADCAST_ADDR_TYPE)
+	|| (orig_addr->addr_type == IPMI_IPMB_ADDR_TYPE))
+    {
+	ipmi_ipmb_addr_t *ipmb2 = (ipmi_ipmb_addr_t *) orig_addr;
+	chan = ipmb2->channel;
+    } else
+	chan = 0;
+
     if ((tmsg[5] == IPMI_SEND_MSG_CMD)
 	&& ((tmsg[1] >> 2) == (IPMI_APP_NETFN | 1)))
     {
@@ -222,16 +235,27 @@ ipmi_handle_recv(ipmi_con_t    *ipmi,
 	    msg->cmd = orig_msg->cmd;
 	    msg->data = tmsg + 6;
 	    msg->data_len = 1;
-	    if (ipmi->handle_send_rsp_err) {
-		ipmi->handle_send_rsp_err(ipmi, msg);
-	    }
+	    to_ret = -1;
 	} else {
 	    if (data_len < 15)
 		/* The response to a send message was not carrying the
 		   payload. */
 		return EINVAL;
 
-	    if (tmsg[10] == ipmi->ipmb_addr) {
+	    if ((orig_msg->cmd == IPMI_SEND_MSG_CMD)
+		&& (orig_msg->netfn == IPMI_APP_NETFN))
+	    {
+		/* Boy, I hate IPMI message routing.  If the original
+		   command from the user was a send message, then
+		   assume this was the response the user was looking
+		   for.  This means that you can't route send message
+		   responses through a connection that supports
+		   returning the message data in the send message
+		   response, but that's wrong, anyway. */
+		goto handle_normal_msg;
+	    }
+
+	    if (tmsg[10] == ipmi->ipmb_addr[chan]) {
 		ipmi_system_interface_addr_t *si_addr
 		    = (ipmi_system_interface_addr_t *) addr;
 
@@ -255,61 +279,67 @@ ipmi_handle_recv(ipmi_con_t    *ipmi,
 	    msg->data = tmsg+13;
 	    msg->data_len = data_len - 15;
 	}
-    } else if ((orig_addr->addr_type != IPMI_SYSTEM_INTERFACE_ADDR_TYPE)
-	       && (((ipmi->hacks & IPMI_CONN_HACK_20_AS_MAIN_ADDR)
-		    && (tmsg[3] == 0x20))
-		   || ((! (ipmi->hacks & IPMI_CONN_HACK_20_AS_MAIN_ADDR))
-		       && (tmsg[3] == ipmi->ipmb_addr))))
-    {
-        /* In some cases, a message from the IPMB looks like it came
-	   from the BMC itself, IMHO a misinterpretation of the
-	   errata.  IPMIv1_5_rev1_1_0926 markup, section 6.12.4,
-	   didn't clear things up at all.  Some manufacturers have
-	   interpreted it this way, but IMHO it is incorrect. */
-        memcpy(addr, orig_addr, orig_addr_len);
-        addr_len = orig_addr_len;
-	if (addr->addr_type == IPMI_IPMB_BROADCAST_ADDR_TYPE)
-	    addr->addr_type = IPMI_IPMB_ADDR_TYPE;
-        msg->netfn = tmsg[1] >> 2;
-        msg->cmd = tmsg[5];
-        msg->data = tmsg+6;
-        msg->data_len = data_len - 7;
     } else {
-	/* It's not encapsulated in a send message response. */
-
-	if (((ipmi->hacks & IPMI_CONN_HACK_20_AS_MAIN_ADDR)
-	     && (tmsg[3] == 0x20))
-	    || ((!(ipmi->hacks & IPMI_CONN_HACK_20_AS_MAIN_ADDR))
-		&& (tmsg[3] == ipmi->ipmb_addr)))
+    handle_normal_msg:
+	if ((orig_addr->addr_type != IPMI_SYSTEM_INTERFACE_ADDR_TYPE)
+	    && (((ipmi->hacks & IPMI_CONN_HACK_20_AS_MAIN_ADDR)
+		 && (tmsg[3] == 0x20))
+		|| ((! (ipmi->hacks & IPMI_CONN_HACK_20_AS_MAIN_ADDR))
+		    && (tmsg[3] == ipmi->ipmb_addr[chan]))))
 	{
-	    ipmi_system_interface_addr_t *si_addr
-		= (ipmi_system_interface_addr_t *) addr;
-
-	    /* It's directly from the BMC, so it's a system interface
-	       message. */
-	    si_addr->addr_type = IPMI_SYSTEM_INTERFACE_ADDR_TYPE;
-	    si_addr->channel = 0xf;
-	    si_addr->lun = tmsg[4] & 3;
+	    /* In some cases, a message from the IPMB looks like it came
+	       from the BMC itself, IMHO a misinterpretation of the
+	       errata.  IPMIv1_5_rev1_1_0926 markup, section 6.12.4,
+	       didn't clear things up at all.  Some manufacturers have
+	       interpreted it this way, but IMHO it is incorrect. */
+	    /* That said, this is the way things are.  This appears to be
+	       the correct interpretation of the spec, even though it's
+	       crazy. */
+	    memcpy(addr, orig_addr, orig_addr_len);
+	    addr_len = orig_addr_len;
+	    if (addr->addr_type == IPMI_IPMB_BROADCAST_ADDR_TYPE)
+		addr->addr_type = IPMI_IPMB_ADDR_TYPE;
+	    msg->netfn = tmsg[1] >> 2;
+	    msg->cmd = tmsg[5];
+	    msg->data = tmsg+6;
+	    msg->data_len = data_len - 7;
 	} else {
-	    ipmi_ipmb_addr_t *ipmb_addr	= (ipmi_ipmb_addr_t *) addr;
-	    ipmi_ipmb_addr_t *ipmb2 = (ipmi_ipmb_addr_t *) orig_addr;
+	    /* It's not encapsulated in a send message response. */
 
-	    /* A message from the IPMB. */
-	    ipmb_addr->addr_type = IPMI_IPMB_ADDR_TYPE;
-	    /* This is a hack, but the channel does not come back in the
-	       message.  So we use the channel from the original
-	       instead. */
-	    ipmb_addr->channel = ipmb2->channel;
-	    ipmb_addr->slave_addr = tmsg[3];
-	    ipmb_addr->lun = tmsg[4] & 0x3;
+	    if (((ipmi->hacks & IPMI_CONN_HACK_20_AS_MAIN_ADDR)
+		 && (tmsg[3] == 0x20))
+		|| ((!(ipmi->hacks & IPMI_CONN_HACK_20_AS_MAIN_ADDR))
+		    && (tmsg[3] == ipmi->ipmb_addr[chan])))
+	    {
+		ipmi_system_interface_addr_t *si_addr
+		    = (ipmi_system_interface_addr_t *) addr;
+
+		/* It's directly from the BMC, so it's a system interface
+		   message. */
+		si_addr->addr_type = IPMI_SYSTEM_INTERFACE_ADDR_TYPE;
+		si_addr->channel = 0xf;
+		si_addr->lun = tmsg[4] & 3;
+	    } else {
+		ipmi_ipmb_addr_t *ipmb_addr	= (ipmi_ipmb_addr_t *) addr;
+		ipmi_ipmb_addr_t *ipmb2 = (ipmi_ipmb_addr_t *) orig_addr;
+
+		/* A message from the IPMB. */
+		ipmb_addr->addr_type = IPMI_IPMB_ADDR_TYPE;
+		/* This is a hack, but the channel does not come
+		   back in the message.  So we use the channel
+		   from the original instead. */
+		ipmb_addr->channel = ipmb2->channel;
+		ipmb_addr->slave_addr = tmsg[3];
+		ipmb_addr->lun = tmsg[4] & 0x3;
+	    }
+
+	    msg->netfn = tmsg[1] >> 2;
+	    msg->cmd = tmsg[5];
+	    addr_len = sizeof(ipmi_system_interface_addr_t);
+	    msg->data = tmsg+6;
+	    msg->data_len = data_len - 6;
+	    msg->data_len--; /* Remove the checksum */
 	}
-
-	msg->netfn = tmsg[1] >> 2;
-	msg->cmd = tmsg[5];
-	addr_len = sizeof(ipmi_system_interface_addr_t);
-	msg->data = tmsg+6;
-	msg->data_len = data_len - 6;
-	msg->data_len--; /* Remove the checksum */
     }
     
     /* Convert broadcast addresses to regular IPMB addresses, since
@@ -367,7 +397,7 @@ ipmi_handle_recv(ipmi_con_t    *ipmi,
         ipmi_log(IPMI_LOG_DEBUG_END, " ");
     }
 
-    return 0;
+    return to_ret;
 }
 
 static void
