@@ -624,7 +624,7 @@ start_config_fetch_cb(ipmi_mc_t *mc, void *cb_data)
     return;
 }
 
-static void
+static int
 start_config_fetch(void *cb_data, int shutdown)
 {
     pef_fetch_handler_t *elem = cb_data;
@@ -636,7 +636,7 @@ start_config_fetch(void *cb_data, int shutdown)
 		 "PEF was destroyed while an operation was in progress");
 	pef_lock(elem->pef);
 	fetch_complete(elem->pef, ECANCELED, elem);
-	return;
+	return OPQ_HANDLER_STARTED;
     }
 
     /* The read lock must be claimed before the pef lock to avoid
@@ -647,6 +647,7 @@ start_config_fetch(void *cb_data, int shutdown)
 	pef_lock(elem->pef);
 	fetch_complete(elem->pef, rv, elem);
     }
+    return OPQ_HANDLER_STARTED;
 }
 
 int
@@ -785,7 +786,7 @@ start_config_set_cb(ipmi_mc_t *mc, void *cb_data)
     return;
 }
 
-static void
+static int
 start_config_set(void *cb_data, int shutdown)
 {
     pef_set_handler_t *elem = cb_data;
@@ -797,7 +798,7 @@ start_config_set(void *cb_data, int shutdown)
 		 "PEF was destroyed while an operation was in progress");
 	pef_lock(elem->pef);
 	set_complete(elem->pef, ECANCELED, elem);
-	return;
+	return OPQ_HANDLER_STARTED;
     }
 
     /* The read lock must be claimed before the pef lock to avoid
@@ -808,6 +809,7 @@ start_config_set(void *cb_data, int shutdown)
 	pef_lock(elem->pef);
 	set_complete(elem->pef, rv, elem);
     }
+    return OPQ_HANDLER_STARTED;
 }
 
 int
@@ -1041,7 +1043,7 @@ struct ipmi_pef_config_s
 
     unsigned char num_alert_strings;	/* Number of alert strings */
     ipmi_ask_t	  *asks;		/* Alert String Key Table */
-    unsigned char **alert_strings;	/* Alert strings */
+    char          **alert_strings;	/* Alert strings */
 };
 
 
@@ -1465,10 +1467,10 @@ static void sask(ipmi_pef_config_t *pefc, pefparms_t *lp, unsigned char *data,
 static int gas(ipmi_pef_config_t *pefc, pefparms_t *lp, int err,
 	       unsigned char *data, unsigned int data_len)
 {
-    int           pos;
-    unsigned char **t;
-    unsigned char *s1, *s2;
-    int           len;
+    int  pos;
+    char **t;
+    char *s1, *s2;
+    int  len;
 
     data++; /* Skip the revision byte. */
     data_len--;
@@ -1491,7 +1493,7 @@ static int gas(ipmi_pef_config_t *pefc, pefparms_t *lp, int err,
        zeros, either. */
     s1 = *t;
     if (s1)
-	len = strlen(s1);
+	len = strlen((char *) s1);
     else
 	len = 0;
     s2 = ipmi_mem_alloc(len + data_len + 1);
@@ -1514,10 +1516,10 @@ static int gas(ipmi_pef_config_t *pefc, pefparms_t *lp, int err,
 static void sas(ipmi_pef_config_t *pefc, pefparms_t *lp, unsigned char *data,
 		unsigned int *data_len)
 {
-    int           pos = data[0] & 0x7f;
-    int           block = data[1];
-    unsigned char *t = pefc->alert_strings[pos];
-    int           len;
+    int  pos = data[0] & 0x7f;
+    int  block = data[1];
+    char *t = pefc->alert_strings[pos];
+    int  len;
 
     if (!t) {
 	data[2] = '\0';
@@ -1585,7 +1587,7 @@ got_parm(ipmi_pef_t     *pef,
     pefparms_t        *lp = &(pefparms[pefc->curr_parm]);
 
     /* Check the length, and don't forget the revision byte must be added. */
-    if ((!err) && (data_len < (lp->length+1))) {
+    if ((!err) && (data_len < (unsigned int) (lp->length+1))) {
 	ipmi_log(IPMI_LOG_ERR_INFO,
 		 "ipmi_pefparm_got_parm:"
 		 " Invalid data length on parm %d was %d, should have been %d",
@@ -1766,7 +1768,7 @@ lock_done(ipmi_pef_t *pef,
 	pefc->lock_supported = 0;
     } else if (err == IPMI_IPMI_ERR_VAL(0x81)) {
 	/* Someone else has the lock, return EAGAIN. */
-	pefc->done(pef, err, NULL, pefc->cb_data);
+	pefc->done(pef, EAGAIN, NULL, pefc->cb_data);
 	ipmi_pef_free_config(pefc);
 	pef_put(pef);
 	return;
@@ -1847,7 +1849,8 @@ set_clear(ipmi_pef_t *pef,
 
     if (pefc->err)
 	err = pefc->err;
-    pefc->set_done(pef, err, pefc->cb_data);
+    if (pefc->set_done)
+	pefc->set_done(pef, err, pefc->cb_data);
     ipmi_pef_free_config(pefc);
     pef_put(pef);
 }
@@ -1987,7 +1990,7 @@ set_done(ipmi_pef_t *pef,
     if (pefc->curr_parm == IPMI_PEFPARM_ALERT_STRING) {
 	/* Special handling for blocks of an alert string */
 	if ((length < 18) ||
-	    (memchr(data+2, '\0', length)))
+	    (memchr(data+2, '\0', length-2)))
 	{
 	    /* End of string, either a subsize-block or a nil
 	       character in the data. */
@@ -2419,7 +2422,7 @@ ipmi_pefconfig_get_alert_string(ipmi_pef_config_t *pefc, unsigned int sel,
     *len = rlen;
     if (rlen > olen)
 	return EBADF;
-    strcpy(val, pefc->alert_strings[sel]);
+    strcpy((char *) val, pefc->alert_strings[sel]);
     return 0;
 }
 
@@ -2431,10 +2434,10 @@ ipmi_pefconfig_set_alert_string(ipmi_pef_config_t *pefc, unsigned int sel,
 
     if (sel >= pefc->num_alert_strings)
 	return EINVAL;
-    s = pefc->alert_strings[sel];
+    s = (unsigned char *) pefc->alert_strings[sel];
     pefc->alert_strings[sel] = ipmi_strdup((char *) val);
     if (! pefc->alert_strings[sel]) {
-	pefc->alert_strings[sel] = s;
+	pefc->alert_strings[sel] = (char *) s;
 	return ENOMEM;
     }
     if (s)
@@ -2446,7 +2449,7 @@ ipmi_pefconfig_set_alert_string(ipmi_pef_config_t *pefc, unsigned int sel,
 typedef struct pefparm_gendata_s
 {
     enum ipmi_pefconf_val_type_e datatype;
-    unsigned char *fname;
+    char *fname;
 
     union {
 	struct {
@@ -2581,6 +2584,19 @@ static pefparm_gendata_t gdata[] =
 };
 #define NUM_GDATA_ENTRIES (sizeof(gdata) / sizeof(pefparm_gendata_t))
 
+int
+ipmi_pefconfig_enum_val(unsigned int parm, int val, int *nval,
+			const char **sval)
+{
+    return ENOSYS;
+}
+
+int
+ipmi_pefconfig_enum_idx(unsigned int parm, int idx, const char **sval)
+{
+    return ENOSYS;
+}
+
 int
 ipmi_pefconfig_get_val(ipmi_pef_config_t *pefc,
 		       unsigned int      parm,
@@ -2707,7 +2723,7 @@ ipmi_pefconfig_set_val(ipmi_pef_config_t *pefc,
 
     if (gdata[parm].iv_cnt) {
 	count = gdata[parm].iv_cnt(pefc);
-	if (index >= count)
+	if (index >= (int) count)
 	    return E2BIG;
     }
 
@@ -2758,7 +2774,7 @@ ipmi_pefconfig_data_free(void *data)
 unsigned int
 ipmi_pefconfig_str_to_parm(char *name)
 {
-    int i;
+    unsigned int i;
     for (i=0; i<NUM_GDATA_ENTRIES; i++) {
 	if (strcmp(name, gdata[i].fname) == 0)
 	    return i;
